/* 
	Corsix's N Level Generator
    Copyright (C) 2007 Peter Cawley

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "LuaGenerator.h"
extern "C"
{
#include <lauxlib.h>
#include <lualib.h>
}
#include <string.h>

LuaGenerator::LuaGenerator()
{
	m_pL = luaL_newstate();
	m_pWorkerThread = lua_newthread(m_pL);
	lua_setfield(m_pL, LUA_REGISTRYINDEX, "m_pWorkerThread");
	m_sLastError = 0;
}

LuaGenerator::~LuaGenerator()
{
	delete[] m_sLastError;
	lua_close(m_pL);
}

void LuaGenerator::_setLastError(const char* sWhat, const char* sDetail)
{
	delete[] m_sLastError;
	m_sLastError = new wchar_t[strlen(sWhat) + strlen(sDetail) + 1];
	wchar_t* p = m_sLastError;
	while(*sWhat)
	{
		*p = *sWhat;
		++sWhat;
		++p;
	}
	while(*sDetail)
	{
		*p = *sDetail;
		++sDetail;
		++p;
	}
	*p = 0;
}

bool LuaGenerator::_catchLoadErr(lua_State *L, int iCode)
{
	switch(iCode)
	{
	case 0:
		return true;
	case LUA_ERRSYNTAX:
		_setLastError("Syntax error during load:\n", lua_tostring(L, -1));
		return false;
	case LUA_ERRMEM:
		_setLastError("Memory error during load:\n", lua_tostring(L, -1));
		return false;
	case LUA_ERRFILE:
		_setLastError("File error during load:\n", lua_tostring(L, -1));
		return false;
	default:
		_setLastError("Unknown error during load:\n", lua_tostring(L, -1));
		return false;
	}
}

bool LuaGenerator::_catchPCallErr(lua_State *L, int iCode)
{
	switch(iCode)
	{
	case 0:
		return true;
	case LUA_ERRRUN:
		_setLastError("Runtime error:\n", lua_tostring(L, -1));
		return false;
	case LUA_ERRMEM:
		_setLastError("Memory error:\n", lua_tostring(L, -1));
		return false;
	case LUA_ERRERR:
		_setLastError("Error handling error:\n", lua_tostring(L, -1));
		return false;
	default:
		_setLastError("Unknown error:\n", lua_tostring(L, -1));
		return false;
	}
}

const wchar_t* LuaGenerator::getErrorDetail()
{
	return m_sLastError ? m_sLastError : L"";
}

static int Lua_setOverlay(lua_State *L)
{
	frmMain* p = (frmMain*)lua_touserdata(L, lua_upvalueindex(1));
	p->getTerrain().setOverlay(lua_tostring(L, 1));
	return 0;
}

static int Lua_setName(lua_State *L)
{
	frmMain* p = (frmMain*)lua_touserdata(L, lua_upvalueindex(1));
	p->getNameTextbox()->SetValue( wxString(lua_tostring(L, 1), wxConvUTF8) );
	return 0;
}

static int Lua_setAuthor(lua_State *L)
{
	frmMain* p = (frmMain*)lua_touserdata(L, lua_upvalueindex(1));
	p->getAuthorTextbox()->SetValue( wxString(lua_tostring(L, 1), wxConvUTF8) );
	return 0;
}

void LuaGenerator::setSpeed(int iSpeed)
{
	lua_pushinteger(m_pL, iSpeed);
	lua_setglobal(m_pL, "FEEDBACK_SPEED");
}

bool LuaGenerator::init(const char* sFilename, frmMain* pDestination)
{
	lua_settop(m_pL, 0);
	lua_settop(m_pWorkerThread, 0);

	const char* sErrCatch = "errorcatch = function(e,t) return tostring(e) .. [[\n\n]] .. ((t and debug.traceback(t)) or debug.traceback()) end";
	if(!_catchLoadErr(m_pL, luaL_loadbuffer(m_pL, sErrCatch, strlen(sErrCatch), "error catcher"))
	|| !_catchPCallErr(m_pL, lua_pcall(m_pL, 0, 0, 0)) )
		return false;

	luaL_openlibs(m_pL);

	lua_pushlightuserdata(m_pL, (void*)pDestination);
	lua_pushcclosure(m_pL, Lua_setOverlay, 1);
	lua_setglobal(m_pL, "setOverlay");

	lua_pushlightuserdata(m_pL, (void*)pDestination);
	lua_pushcclosure(m_pL, Lua_setName, 1);
	lua_setglobal(m_pL, "setName");

	lua_pushlightuserdata(m_pL, (void*)pDestination);
	lua_pushcclosure(m_pL, Lua_setAuthor, 1);
	lua_setglobal(m_pL, "setAuthor");

	// Number metatable
	lua_settop(m_pL, 0);
	lua_pushnumber(m_pL, 0);
	lua_createtable(m_pL, 0, 1);
	lua_getglobal(m_pL, "math");
	lua_setfield(m_pL, 2, "__index");
	lua_setmetatable(m_pL, 1);
	lua_pop(m_pL, 1);

	lua_getglobal(m_pL, "errorcatch");

	if(!_catchLoadErr(m_pL, luaL_loadfile(m_pL, sFilename))
	|| !_catchPCallErr(m_pL, lua_pcall(m_pL, 0, 0, 1)) )
		return false;

	return true;
}

wxStopWatch oInfiniteTimer;
void Infinite_Check_Hook(lua_State *L, lua_Debug *ar)
{
	if(oInfiniteTimer.Time() > 7291)
	{
		oInfiniteTimer.Start(0);
		if(::wxMessageBox(wxT("The generator script is taking a long time; it may be stuck in an infinite loop.\nDo you want to abort it?"),
			wxT("Level Generator"), wxYES_NO | wxICON_QUESTION) == wxYES)
		{
			lua_pushlightuserdata(L, 0);
			lua_error(L);
			return;
		}
	}
}

bool LuaGenerator::generate(frmMain* pDestination)
{
	oInfiniteTimer.Start(0);
	lua_sethook(m_pWorkerThread, Infinite_Check_Hook, LUA_MASKCOUNT, 10000);

	lua_settop(m_pWorkerThread, 0);
	lua_getglobal(m_pL, "generate");
	lua_xmove(m_pL, m_pWorkerThread, 1);
	int iStatus = lua_resume(m_pWorkerThread, 0);
	while(iStatus == LUA_YIELD)
	{
		oInfiniteTimer.Start(0);
		if(lua_type(m_pWorkerThread, -1) == LUA_TSTRING)
		{
			pDestination->setData(lua_tostring(m_pWorkerThread, -1));
		}
		lua_settop(m_pWorkerThread, 0);
		iStatus = lua_resume(m_pWorkerThread, 0);
	}
	if(iStatus != 0)
	{
		if(lua_type(m_pWorkerThread, -1) == LUA_TLIGHTUSERDATA)
		{
			_setLastError("Script aborted","");
			return false;
		}

		lua_getglobal(m_pL, "errorcatch");
		lua_pushvalue(m_pWorkerThread, -1);
		lua_xmove(m_pWorkerThread, m_pL, 1);
		lua_getfield(m_pL, LUA_REGISTRYINDEX, "m_pWorkerThread");
		if(lua_pcall(m_pL, 2, 1, 0) == 0)
		{
			lua_pop(m_pWorkerThread, 1);
			lua_xmove(m_pL, m_pWorkerThread, 1);
		}
		else
		{
			//lua_pop(m_pWorkerThread, 1);
			//lua_xmove(m_pL, m_pWorkerThread, 1);

			lua_pop(m_pL, 1);
			iStatus = LUA_ERRERR;
		}
		return _catchPCallErr(m_pWorkerThread, iStatus);
	}
	if(lua_type(m_pWorkerThread, -1) == LUA_TSTRING)
	{
		pDestination->setData(lua_tostring(m_pWorkerThread, -1));
	}
	return true;
}