/* 
	Corsix's N Level Generator
    Copyright (C) 2007 Peter Cawley

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Terrain.h"
#include "wx/dcmemory.h"

#define STATIC_ARRAY_SIZE(a) (sizeof((a))/sizeof(*(a)))

static int dist_sqr(int x1, int y1, int x2, int y2)
{
	x1 -= x2;
	y1 -= y2;
	return x1 * x1 + y1 * y1;
}

struct colour_t
{
	unsigned char r,g,b;
};

static void CanvasToTilesheet(unsigned char* pCanvas, int iTileNum, bool bMakeCompliment, unsigned char* pTilesheet, colour_t* pPalette)
{
	const int iOutputBPP = 3;
	const int iOutputPitch = Terrain::k_iCellSize * iOutputBPP;
	const int iInputRow0 = Terrain::k_iCellSize *  0;
	const int iInputRow1 = Terrain::k_iCellSize *  4;
	const int iInputRow2 = Terrain::k_iCellSize *  8;
	const int iInputRow3 = Terrain::k_iCellSize * 12;

	iTileNum = iTileNum * Terrain::k_iCellSize * iOutputPitch;
	for(int y = 0; y < Terrain::k_iCellSize; ++y)
	{
		for(int x = 0; x < Terrain::k_iCellSize; ++x)
		{
			int iOutputOffset = iTileNum + y * iOutputPitch + x * iOutputBPP;
			int iInputOffset  = y * Terrain::k_iCellSize * 16 + x * 4;
			int iPaletteIndex = 
				pCanvas[iInputOffset + iInputRow0 + 0] +
				pCanvas[iInputOffset + iInputRow0 + 1] +
				pCanvas[iInputOffset + iInputRow0 + 2] +
				pCanvas[iInputOffset + iInputRow0 + 3] +
				pCanvas[iInputOffset + iInputRow1 + 0] +
				pCanvas[iInputOffset + iInputRow1 + 1] +
				pCanvas[iInputOffset + iInputRow1 + 2] +
				pCanvas[iInputOffset + iInputRow1 + 3] +
				pCanvas[iInputOffset + iInputRow2 + 0] +
				pCanvas[iInputOffset + iInputRow2 + 1] +
				pCanvas[iInputOffset + iInputRow2 + 2] +
				pCanvas[iInputOffset + iInputRow2 + 3] +
				pCanvas[iInputOffset + iInputRow3 + 0] +
				pCanvas[iInputOffset + iInputRow3 + 1] +
				pCanvas[iInputOffset + iInputRow3 + 2] +
				pCanvas[iInputOffset + iInputRow3 + 3] ;
			pTilesheet[iOutputOffset + 0] = pPalette[iPaletteIndex].r;
			pTilesheet[iOutputOffset + 1] = pPalette[iPaletteIndex].g;
			pTilesheet[iOutputOffset + 2] = pPalette[iPaletteIndex].b;

			if(bMakeCompliment)
			{
				iPaletteIndex = 16 - iPaletteIndex;
				iOutputOffset += iOutputPitch * Terrain::k_iCellSize;

				pTilesheet[iOutputOffset + 0] = pPalette[iPaletteIndex].r;
				pTilesheet[iOutputOffset + 1] = pPalette[iPaletteIndex].g;
				pTilesheet[iOutputOffset + 2] = pPalette[iPaletteIndex].b;
			}
		}
	}
}

Terrain::Terrain()
{
	m_pCellTiles = new int[k_iWidth * k_iHeight];
	m_pCellTilesOverlay = new wxString[k_iWidth * k_iHeight];

	for(int y = 0; y < k_iHeight; ++y)
	{
		for(int x = 0; x < k_iWidth; ++x)
		{
			if(x == 0 || x == (k_iWidth - 1) || y == 0 || y == (k_iHeight - 1))
			{
				m_pCellTiles[y * k_iWidth + x] = 32;
			}
			else
			{
				m_pCellTiles[y * k_iWidth + x] = 33;
			}
		}
	}
}

Terrain::~Terrain()
{
	for(std::vector<_object_t*>::iterator itr = m_vObjects.begin(); itr != m_vObjects.end(); ++itr)
		delete *itr;
	delete[] m_pCellTilesOverlay;
	delete[] m_pCellTiles;
}

void Terrain::setOverlay(const char* sOverlayData)
{
	wchar_t s[2];
	s[1] = 0;
	for(int x = 1; x < (k_iWidth - 1); ++x)
	{
		for(int y = 1; y < (k_iHeight - 1); ++y)
		{
			s[0] = *sOverlayData;
			if(s[0] == 0)
				s[0] = ' ';
			else
				++sOverlayData;
			m_pCellTilesOverlay[y * k_iWidth + x] = s;
		}
	}
}

void Terrain::load(const char* sMapData)
{
	for(std::vector<_object_t*>::iterator itr = m_vObjects.begin(); itr != m_vObjects.end(); ++itr)
		delete *itr;
	m_vObjects.clear();

	const char* sPolygonNames = ";9:8<6=7OQNP3524GM>D@BIK?EFLHJAC10";

	int iNameToId[256];
	for(int i = 0; i < 256; ++i)
		iNameToId[i] = -1;
	iNameToId[0] = -2;

	for(int i = 0; sPolygonNames[i]; ++i)
		iNameToId[ sPolygonNames[i] ] = i;

	for(int x = 1; x < (k_iWidth - 1); ++x)
	{
		for(int y = 1; y < (k_iHeight - 1); ++y)
		{
			int iPolygon;
			while( (iPolygon = iNameToId[*sMapData]) == -1) ++sMapData;
			if(iPolygon < 0)
				return;
			else
				m_pCellTiles[y * k_iWidth + x] = iPolygon;
			++sMapData;
		}
	}

	if(*sMapData == '|')
	{
		bool bLastObject = false;
		const char* sObjEnd = sMapData;
		const char* sObjStart;
		do
		{
			_object_t *pObj = 0;
			sObjStart = sObjEnd + 1;
			size_t iObjLen;
			sObjEnd = strchr(sObjStart, '!');
			if(sObjEnd == 0)
			{
				iObjLen = strlen(sObjStart);
				bLastObject = true;
			}
			else
			{
				iObjLen = sObjEnd - sObjStart;
			}

			switch(*sObjStart)
			{
			case '0':
				pObj = new _gold_t;
				break;
			case '1':
				switch(sObjStart[1])
				{
				case '^':
					pObj = new _bounce_t;
					break;
				case '0':
					pObj = new _rocket_t;
					break;
				case '1':
					pObj = new _exit_t;
					break;
				case '2':
					pObj = new _mine_t;
					break;
				}
				break;
			case '2':
				pObj = new _launchpad_t;
				break;
			case '3':
				pObj = new _gauss_t;
				break;
			case '5':
				pObj = new _ninja_t;
				break;
			case '6':
				pObj = new _drone_t;
				break;
			case '8':
				pObj = new _thuwmp_t;
				break;
			case '9':
				pObj = new _door_t;
				break;
			}

			if(pObj != 0)
			{
				pObj->load(sObjStart, iObjLen);
				m_vObjects.push_back(pObj);
			}

		} while(!bLastObject);
	}
}

bool Terrain::_ninja_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	return oDest.Blit(x - 9, y - 10, 18, 20, pObjectSheet, 0, 0, wxCOPY, true);
}

bool Terrain::_launchpad_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	return oDest.Blit(x - 8, y - 6, 16, 6, pObjectSheet, 16, 38, wxCOPY, true);
}

bool Terrain::_gold_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	return oDest.Blit(x - 3, y - 3, 7, 7, pObjectSheet, 18, 0, wxCOPY, true);
}

bool Terrain::_mine_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	return oDest.Blit(x - 4, y - 4, 8, 8, pObjectSheet, 18, 8, wxCOPY, true);
}

bool Terrain::_rocket_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	return oDest.Blit(x - 7, y - 6, 14, 13, pObjectSheet, 24, 24, wxCOPY, true);
}

bool Terrain::_gauss_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	return oDest.Blit(x - 7, y - 6, 14, 13, pObjectSheet, 0, 33, wxCOPY, true);
}

bool Terrain::_bounce_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	return oDest.Blit(x - 10, y - 10, 20, 20, pObjectSheet, 12, 84, wxCOPY, true);
}

bool Terrain::_thuwmp_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	return oDest.Blit(x - 9, y - 9, 19, 19, pObjectSheet, 0 + (d / 2) * 19, 46 + (d % 2) * 19, wxCOPY, true);
}

bool Terrain::_exit_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	return oDest.Blit(x - 13, y - 12, 26, 25, pObjectSheet, 0, 106, wxCOPY, true)
		&& oDest.Blit(sx - 7, sy - 4, 13,  8, pObjectSheet,26, 123, wxCOPY, true);
}

bool Terrain::_door_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	if(!oDest.Blit(sx - 4, sy - 4, 8,  8, pObjectSheet, 0, 98, wxCOPY, true))
		return false;

	if(o == 0)
		return oDest.Blit( (x + 1 + o1) * 24 - 2, y * 24 - 1, 5, 26, pObjectSheet, 38, 58, wxCOPY, true);
	else
		return oDest.Blit( x * 24, (y + 1 + o2) * 24 - 2, 24, 5, pObjectSheet, 40, 126, wxCOPY, true);
}

bool Terrain::_drone_t::render(wxDC& oDest, wxDC* pObjectSheet)
{
	return oDest.Blit(x - 12, y - 15, 24, 24, pObjectSheet, dir * 24, type * 24 + seeker * 72, wxCOPY, true);
}

static int readNum(const char* &s)
{
	bool bNeg = false;
	int i = 0;
	if(*s == '-')
	{
		++s;
		bNeg = true;
	}
	while(*s >= '0' && *s <= '9')
	{
		i *= 10;
		i += (*s - '0');
		++s;
	}
	if(bNeg)
		i = -i;
	return i;
}

static double readFloat(const char* &s)
{
	double f = atof(s);
	bool bInDecimals = false;
	if(*s == '-') ++s;
	while( (*s >= '0' && *s <= '9') || (!bInDecimals && *s == '.' && (bInDecimals = true)) ) ++s;
	return f;
}

void Terrain::_ninja_t::load(const char* sData, size_t iLen)
{
	sData += 2; // "5^"
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
}

void Terrain::_launchpad_t::load(const char* sData, size_t iLen)
{
	sData += 2; // "2^"
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
	sData += 1; // ","
	xt = readFloat(sData); // x thrust
	sData += 1; // ","
	yt = readFloat(sData); // y thrust
}

void Terrain::_gold_t::load(const char* sData, size_t iLen)
{
	sData += 2; // "0^"
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
}

void Terrain::_thuwmp_t::load(const char* sData, size_t iLen)
{
	sData += 2; // "8^"
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
	sData += 1; // ","
	d = readNum(sData); // d
}

void Terrain::_mine_t::load(const char* sData, size_t iLen)
{
	sData += 3; // "12^"
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
}

void Terrain::_gauss_t::load(const char* sData, size_t iLen)
{
	sData += 2; // "3^"
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
}

void Terrain::_bounce_t::load(const char* sData, size_t iLen)
{
	sData += 2; // "1^"
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
}

void Terrain::_rocket_t::load(const char* sData, size_t iLen)
{
	sData += 3; // "10^"
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
}

void Terrain::_exit_t::load(const char* sData, size_t iLen)
{
	sData += 3; // "11^"
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
	sData += 1; // ","
	sx = readNum(sData); // sx
	sData += 1; // ","
	sy = readNum(sData); // sy
}

void Terrain::_door_t::load(const char* sData, size_t iLen)
{
	sData += 2; // "9^"
	sx = readNum(sData); // sx
	sData += 1; // ","
	sy = readNum(sData); // sy
	sData += 1; // ","
	o = readNum(sData); // orient
	sData += 1; // ","
	u1 = readNum(sData); // unknown
	sData += 1; // ","
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
	sData += 1; // ","
	u2 = readNum(sData); // unknown
	sData += 1; // ","
	o1 = readNum(sData); // orient 1
	sData += 1; // ","
	o2 = readNum(sData); // orient 2
}

void Terrain::_drone_t::load(const char* sData, size_t iLen)
{
	sData += 2; // "6^"
	x = readNum(sData); // x
	sData += 1; // ","
	y = readNum(sData); // y
	sData += 1; // ","
	algo = readNum(sData); // algo
	sData += 1; // ","
	seeker = readNum(sData); // seeker
	sData += 1; // ","
	type = readNum(sData); // type
	sData += 1; // ","
	dir = readNum(sData); // dir
}

int Terrain::_object_t::getSheetIndex() {return 0;}
int Terrain::_drone_t::getSheetIndex() {return 1;}


bool Terrain::renderOnto(wxDC& oDest)
{
	wxMemoryDC oSrc;
	oSrc.SelectObjectAsSource(m_bmpTiles);

	wxMemoryDC oSrcObjects[2];
	oSrcObjects[0].SelectObjectAsSource(m_bmpObjects);
	oSrcObjects[1].SelectObjectAsSource(m_bmpObjects2);

	for(std::vector<_object_t*>::iterator itr = m_vObjects.begin(); itr != m_vObjects.end(); ++itr)
	{
		if(!(*itr)->render(oDest, (wxMemoryDC*) &oSrcObjects[ (*itr)->getSheetIndex() ] ))
			return false;
	}

	wxCoord iDestY = 0;
	for(int y = 0; y < k_iHeight; ++y)
	{
		wxCoord iDestX = 0;
		for(int x = 0; x < k_iWidth; ++x)
		{
			int iTile = m_pCellTiles[y * k_iWidth + x];
			if(iTile < 33) // don't render the blank cell (as it doesn't exist)
			{
				if(!oDest.Blit(iDestX, iDestY, k_iCellSize, k_iCellSize, &oSrc, 0, k_iCellSize * iTile, wxCOPY, true))
					return false;
			}
			wxSize oMsgLen = oDest.GetTextExtent(m_pCellTilesOverlay[y * k_iWidth + x]);
			oDest.DrawText(m_pCellTilesOverlay[y * k_iWidth + x], iDestX + 12 - oMsgLen.GetWidth() / 2, iDestY + 12 - oMsgLen.GetHeight() / 2);

			iDestX += k_iCellSize;
		}
		iDestY += k_iCellSize;
	}

	return true;
}

bool Terrain::loadObjectImages()
{
	if(!m_bmpObjects.LoadFile(wxT("objects1.png"),wxBITMAP_TYPE_PNG)
	|| !m_bmpObjects2.LoadFile(wxT("objects2.png"),wxBITMAP_TYPE_PNG) )
		return false;
	wxColour oTransparent(202, 202, 208);
	m_bmpObjects .SetMask(new wxMask(m_bmpObjects , oTransparent));
	m_bmpObjects2.SetMask(new wxMask(m_bmpObjects2, oTransparent));
	return true;
}

bool Terrain::generateTileImages()
{
	const int iCellAASize = k_iCellSize * 4;
	const int i4CellSize = iCellAASize - 1;

	unsigned char *pCanvas[17];
	for(int i = 0; i < STATIC_ARRAY_SIZE(pCanvas); ++i)
		pCanvas[i] = new unsigned char[iCellAASize * iCellAASize];

	for(int y = 0; y < iCellAASize; ++y)
	{
		for(int x = 0; x < iCellAASize; ++x)
		{
			int iOffset = y * iCellAASize + x;

			// Curvy pieces
			pCanvas[ 0][iOffset] = dist_sqr(x, y, i4CellSize, i4CellSize) <= (i4CellSize * i4CellSize) ? 1 : 0;
			pCanvas[ 1][iOffset] = dist_sqr(x, y,          0, i4CellSize) <= (i4CellSize * i4CellSize) ? 1 : 0;
			pCanvas[ 2][iOffset] = dist_sqr(x, y, i4CellSize,          0) <= (i4CellSize * i4CellSize) ? 1 : 0;
			pCanvas[ 3][iOffset] = dist_sqr(x, y,          0,          0) <= (i4CellSize * i4CellSize) ? 1 : 0;

			// Rectangular half blocks
			pCanvas[ 4][iOffset] = x >= (iCellAASize / 2) ? 1 : 0;
			pCanvas[ 5][iOffset] = y >= (iCellAASize / 2) ? 1 : 0;

			// 45 degree half blocks
			pCanvas[ 6][iOffset] = (x+y) >= i4CellSize ? 1 : 0;
			pCanvas[ 7][iOffset] =     x <= y          ? 1 : 0;

			// Triangles
			pCanvas[ 8][iOffset] = (y/-2) < (x - iCellAASize    ) ? 1 : 0;
			pCanvas[ 9][iOffset] = (x/ 2) < (y - iCellAASize / 2) ? 1 : 0;
			pCanvas[10][iOffset] = (y* 2) <  x                    ? 1 : 0;
			pCanvas[11][iOffset] = (y/-2) > (x - iCellAASize / 2) ? 1 : 0;
			pCanvas[12][iOffset] = (x/-2) < (y - iCellAASize    ) ? 1 : 0;
			pCanvas[13][iOffset] = (x* 2) <  y                    ? 1 : 0;
			pCanvas[14][iOffset] = (y/ 2) < (x - iCellAASize / 2) ? 1 : 0;
			pCanvas[15][iOffset] = (x/-2) > (y - iCellAASize / 2) ? 1 : 0;

			// Solid block
			pCanvas[16][iOffset] = 1;
		}
	}

	colour_t clrPalette[17];
	int iR = 202;
	int iG = 202;
	int iB = 208;
	for(int n = 0;; ++n)
	{
		clrPalette[n].r = iR;
		clrPalette[n].g = iG;
		clrPalette[n].b = iB;
		if( ++n >= STATIC_ARRAY_SIZE(clrPalette) )
			break;
		iR -= 5;
		iG -= 5;
		iB -= 4;
		clrPalette[n].r = iR;
		clrPalette[n].g = iG;
		clrPalette[n].b = iB;
		iR -= 5;
		iG -= 5;
		iB -= 5;
	}

	unsigned char* pImageData = (unsigned char*)malloc(k_iCellSize * k_iCellSize * 3 * 33);

	for(int i = 0; i < STATIC_ARRAY_SIZE(pCanvas); ++i)
	{
		CanvasToTilesheet(pCanvas[i], i*2, i+1 < STATIC_ARRAY_SIZE(pCanvas), pImageData, &clrPalette[0]);
		delete[] pCanvas[i];
	}

	wxImage imgTiles(k_iCellSize, k_iCellSize * 33, pImageData);
	imgTiles.SetMaskColour(clrPalette[0].r, clrPalette[0].g, clrPalette[0].b);

	m_bmpTiles = wxBitmap(imgTiles);

	return true;
}